<?php class baseobject implements crudable {

	protected static $fields=array();
	protected static $helpTexts=array();
	
	public static function getFieldValidations(){
		return static::$fields; //static, not self - use the child class's.
	}
	public static function getFieldHelpTexts(){
		return static::$helpTexts; //static, not self - use the child class's.
	}
	
	public static $defaultSortOrder='name ASC';
	
	protected static $adminSelect='SELECT SQL_CALC_FOUND_ROWS [[classname]].*,project.name AS projectname, project.isarchived AS isarchived FROM [[classname]], project WHERE project.id=[[classname]].projectid ';
	protected static $normalSelect='SELECT SQL_CALC_FOUND_ROWS [[classname]].*,project.name AS projectname, project.isarchived AS isarchived FROM [[classname]], project WHERE project.id=[[classname]].projectid ';
	
	/**
	 * Return the SQL restricting records to those on which the user has the specified permission
	 * @param string $accessType One of "read", "update", "delete"
	 * @return string
	 */
	protected static function getProjectClause($accessType){
		return database::getProjectClause($accessType);
	}

	/**
	 * Creates a record in the baseobject table and returns its ID.
	 * @param projectId The ID of the project for the new record
	 * @throws NotFoundException if the project does not exist or the user cannot read it
	 * @throws ForbiddenException if the user can read but not create records in the project
	 */
	protected static function createBaseObject($projectId){
		$className=get_called_class();
		if(empty($projectId)){
			throw new BadRequestException('Project ID not provided.');
		}
		if(!session::isAdmin()){
			$projects=session::getProjectPermissions('create');
			if(!in_array($projectId,$projects)){
				session::refreshProjectPermissions();
				$projects=session::getProjectPermissions('create');
			}
			if(!in_array($projectId,$projects)){
				$projects=session::getProjectPermissions('read');
				if(!in_array($projectId,$projects)){
					throw new NotFoundException("Either the project does not exist or you do not have permission to see it");
				} else {
					throw new ForbiddenException("You do not have permission to create in that project");
				}
			}
		}
		$sqlStatement='INSERT INTO baseobject(projectid,objecttype,creator,createtime) VALUES(:projectid,:objecttype,:userid,:now)';
		$now=new DateTime();
		$now->setTimezone(new DateTimeZone('GMT'));
		$params=array(
			':projectid'=>$projectId,
			':objecttype'=>$className,
			':userid'=>session::getUserId(),
			':now'=>$now->format('Y-m-d H:i:s')
		);
		database::query($sqlStatement,$params);
		return database::getLastInsertId();
	}
	
	/**
	 * Returns the record with the specified ID.
	 * @param id The ID of the desired record
	 * @throws NotFoundException if the record does not exist or the user cannot read it
	 */
	public static function getById($id){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$sqlStatement.=' AND [[classname]].id=:id';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
		$params=array(':id'=>$id);
		return database::queryGetOne($sqlStatement,$params);
	}
	
	
	/**
	 * Returns the record with the specified name.
	 * @param name The name of the desired record
	 * @throws NotFoundException if the record does not exist or the user cannot read it
	 */
	public static function getByName($name){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$sqlStatement.=' AND [[classname]].name=:name';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
		$params=array(':name'=>$name);
		return database::queryGetOne($sqlStatement,$params);
	}
	
	public static function getByNameLike($searchTerms,$request=array()){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$searchTerms=strtolower($searchTerms);
		$sqlStatement=str_replace('SELECT ', 'SELECT "[[classname]]" AS objecttype, ', $sqlStatement);
		$sqlStatement.=' AND LOWER([[classname]].name) LIKE :terms ';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
		$sqlStatement.=' ORDER BY INSTR(LOWER([[classname]], :terms2) '; 
		$sqlStatement.=database::getLimitClause($request);
		$params=array(
				':terms'=>'%'.$searchTerms.'%',
				':terms2'=>'%'.$searchTerms.'%'
		);
		return database::queryGetAll($sqlStatement,$params);
	}
	
	public static final function getByNameLikeMultiClass($searchTerms, $request=array()){
		if(!isset($request['classes'])){
			throw new BadRequestException('Specify classes to search for in querystring, e.g., ?classes=type1,type2,type3');
		}
		$searchTerms=strtolower($searchTerms);
		$classes=explode(',', $request['classes']);
		$parts=array();
		$params=array();
		$counter=1;
		foreach($classes as $className){
			if(!class_exists($className)){
				throw new BadRequestException('Class '.$className.' does not exist');
			}
			if(session::isAdmin()){
				$sqlStatement=static::$adminSelect;
			} else {
				$sqlStatement=static::$normalSelect;
			}
			$sqlStatement=str_replace('SQL_CALC_FOUND_ROWS', '', $sqlStatement);
			$sqlStatement=str_replace('SELECT ', 'SELECT "[[classname]]" AS objecttype, ', $sqlStatement);
			$sqlStatement=str_replace('[[classname]].*', '[[classname]].id, [[classname]].name, [[classname]].projectid', $sqlStatement);
			$sqlStatement.=' AND LOWER([[classname]].name) LIKE :terms'.$counter.' ';
			$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
			$sqlStatement.=forward_static_call_array(array($className, 'getProjectClause'), array('read'));
			$parts[]=$sqlStatement;
			$params[':terms'.$counter]='%'.$searchTerms.'%';
			$counter++;
		}
		$sqlStatement='SELECT objecttype,id,name,projectid,projectname FROM (';
		$sqlStatement.=implode(' UNION ALL ',$parts);		
		$sqlStatement.=') AS combinedresults ORDER BY INSTR(LOWER(name), :outerterm), objecttype, LOWER(name)';
		$sqlStatement.=database::getLimitClause($request);
		$params[':outerterm']=$searchTerms;
		$result=database::queryGetAll($sqlStatement,$params);
		$result['searchterms']=$searchTerms;
		return $result;
	}

	/**
	 * Generates the SQL condition for whether objects can be seen in a multi-class search. The default implementation
	 * returns database::getProjectClause("read"), something line ' AND project.id IN(1,3,4) '. Any overriding implementation
	 * must produce similar SQL for the search query to be valid. 
	 * @return string
	 */
	public static function getReadConditionsForMultiClassSearch(){
		return static::getProjectClause('read');
	}
	
	/**
	 * Returns the records with the specified property.
	 * @param key The name of the property
	 * @param value The value of the property
	 * @param request The request object, used for filtering, limit clause, etc.
	 * @throws NotFoundException if the record does not exist or the user cannot read it
	 */
	public static function getByProperty($key,$value,$request=array()){
		$className=get_called_class();
		//verify key in fields array
		$keys=array_keys(self::getFieldValidations());
	
		if(!in_array($key, $keys) &&'projectid'!=$key){
			throw new BadRequestException('Property '.$key.' not recognised on '.$className);
		}
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$sqlStatement.=' AND [[classname]].'.$key.'=:val ';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		//		$sqlStatement='SELECT x.*,project.name AS projectname, project.isarchived AS isarchived FROM '.$className.' AS x, project WHERE project.id=x.projectid AND x.'.$key.'=:val';
		$sqlStatement.=static::getProjectClause('read');
		$sqlStatement.=database::getOrderClause($request, $className);
		$sqlStatement.=database::getLimitClause($request);
		$params=array(':val'=>$value);
		return database::queryGetAll($sqlStatement,$params);
	}	

	/**
	 * Returns the records with the specified property.
	 * @param key The name of the property
	 * @param value The value of the property
	 * @param request The request object, used for filtering, limit clause, etc.
	 * @throws NotFoundException if the record does not exist or the user cannot read it
	 */
	public static function getByProperties($keyValuePairs,$request=array()){
		$className=get_called_class();
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		$keys=array_keys(self::getFieldValidations());
		$params=array();
		foreach($keyValuePairs as $key=>$value){
			if(!in_array($key, $keys) &&'projectid'!=$key){
				throw new BadRequestException('Property '.$key.' not recognised on '.$className);
			}
			$sqlStatement.=' AND [[classname]].'.$key.'=:'.$key.' ';
			$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
			$params[':'.$key]=$value;
		}
		$sqlStatement.=static::getProjectClause('read');
		$sqlStatement.=database::getOrderClause($request, $className);
		$sqlStatement.=database::getLimitClause($request);
		return database::queryGetAll($sqlStatement,$params);
	}
	
	/**
	 * Returns all records, subject to any limits or filters in the request.
	 * @param request The request object, used for filtering, limit clause, etc.
	 * @throws NotFoundException if the record does not exist or the user cannot read it
	 */
	public static function getAll($request=array()){
		$className=get_called_class();
		//verify key in fields array
		if(session::isAdmin()){
			$sqlStatement=static::$adminSelect;
		} else {
			$sqlStatement=static::$normalSelect;
		}
		//$sqlStatement='SELECT x.*,project.name AS projectname, project.isarchived AS isarchived FROM '.$className.' AS x, project WHERE project.id=x.projectid';
		$sqlStatement=str_replace('[[classname]]', $className, $sqlStatement);
		$sqlStatement.=static::getProjectClause('read');
		$sqlStatement.=database::getOrderClause($request, $className);
		$sqlStatement.=database::getLimitClause($request);
		$params=array();
		return database::queryGetAll($sqlStatement,$params);
	}
	
	public static function getFiles($id, $request=array()){
		$files=file::getByProperty('parentid',$id,$request);
		return $files;
	}
	
	public static function getNotes($id, $request=array()){
		$request['pagenumber']='1';
		$request['pagesize']='100000';
		$request['sortby']='entered';
		$notes=note::getByProperty('parentid',$id,$request);
		return $notes;
	}
	
	
	
	/**
	 * Creates a new record
	 *
	 */
	public static function create($request=array(),$className=null){
		if(null==$className){ $className=get_called_class(); }
		$fields=array();
		if(!isset($request['projectid'])){
			throw new BadRequestException('Project ID not provided.');
		}
		$fieldValidations=forward_static_call($className.'::getFieldValidations');
		foreach($fieldValidations as $fname=>$validations){
			if(isset($request[$fname])){
				validator::validate($fname, $request[$fname], $validations);
				$fields[]=$fname;
			} else {
				validator::validate($fname, null, $validations);
			}
		}
		$keys=implode(',', $fields);
		$values=':'.implode(',:',$fields);
		$baseObjectId=self::createBaseObject($request['projectid']);
		$sqlStatement='INSERT INTO '.$className.'(id,projectid,'.$keys.') VALUES(:id,:projectid,'.$values.')';
		$params=array(':id'=>$baseObjectId, ':projectid'=>$request['projectid']);
		foreach($fields as $f){
			if(isset($request[$f])){
				$params[':'.$f]=$request[$f];
			}
		}
		database::query($sqlStatement,$params);
		return array('type'=>$className, 'created'=>self::getById($baseObjectId));
	}
	
	
	/**
	 * Updates one or more fields for a given record. Fields are updated if the field name is found in the fields array of the calling class.
	 * @param ID The ID of the record
	 * @param request Name-value pairs from the request body. 
	 * @return The updated record
	 * @throws BadRequestException if one or more fields is invalid
	 * @throws NotFoundException if record does not exist or user does not have read permission
	 * @throws ForbiddenException if user can read but not update the record
	 */
	public static function update($id,$request=array()){
		$className=get_called_class();
		if(isset($request['_classname'])){ $className=$request['_classname']; }
		if(!preg_match('/^[A-Za-z0-9_-]*$/',$className)){
			throw new BadRequestException('Bad class name got to baseobject::update');
		}
		if(!session::isAdmin()){
			$sqlStatement='SELECT projectid FROM '.$className.',project 
					WHERE '.$className.'.projectid=project.id 
						AND '.$className.'.id=:id '.static::getProjectClause('update');
			$params=array(':id'=>$id);
			$result=database::queryGetOne($sqlStatement,$params);
			if(empty($result)){
				//id doesn't exist, or no update permissions
				$obj=self::getbyid($id); //will throw NotFoundException if not found or no read permissions
				//still here? No exception was thrown above. Object exists, we can read it, we can't update it.
				throw new ForbiddenException("You do not have permission to update that record");	
			}
		}
		//now we can get on with it
		foreach(static::getFieldValidations() as $field=>$validations){
			if(!preg_match('/^[A-Za-z0-9_-]*$/',$field)){
				throw new BadRequestException('Bad field name got to baseobject::update');
			}
			if(isset($request[$field])){
				validator::validate($field,$request[$field],$validations); //will throw exception if invalid
				$params=array(
					':id'=>$id,
					':newvalue'=>$request[$field]
				);			
				database::query('UPDATE '.$className.' SET '.$field.'=:newvalue WHERE id=:id', $params);
			}
		}
		//Set last edit user and time
		$now=new DateTime();
		$now->setTimezone(new DateTimeZone('GMT'));
		$params=array(
			':id'=>$id,
			':lasteditor'=>session::getUserId(),
			':lastedittime'=>$now->format('Y-m-d H:i:s')
		);
		database::query('UPDATE baseobject SET lasteditor=:lasteditor, lastedittime=:lastedittime WHERE id=:id', $params);
		if(isset($request['projectid'])){
			
		}
		return array('updated'=>static::getById($id));
	}
	
	
	/**
	 * Deletes the record with the specified ID, or throws an exception. Note that this does an SQL DELETE of the row in the 
	 * baseobject table ONLY - it relies on the correct ON DELETE CASCADE behaviour being set in the child table definitions.
	 * @param id The database ID of the object to delete
	 * @throws NotFoundException if record does not exist or user does not have read permission
	 * @throws ForbiddenException if the user can read but not delete the record
	 */
	public static function delete($id){
		$sqlStatement='DELETE FROM baseobject WHERE id=:id'.static::getProjectClause('delete');
		$params=array(':id'=>$id);
		$result=database::query($sqlStatement,$params);
		//if affected rows, OK, else
		//get by id
		return array('deleted'=>$id);
	}
	
	/**
	 * Whether the current user can create records in any project.
	 * This default implementation returns true if the user (a) is an administrator, or (b) has create permissions on any project.
	 */
	public static function canCreate(){
		return session::isAdmin() || !empty(session::getCreateProjects());
	}
	
	/**
	 * Whether the current user can update the record with this ID.
	 * This default implementation returns true if the user (a) is an administrator, or (b) has update permissions on the record's project.
	 */
	public static function canUpdate($id){
		if(session::isAdmin()){ return true; }
		$projects=session::getUpdateProjects();
		$item=static::getById($id);
		return in_array($item['projectid'], $projects);
	}
	
	
	
}
?>