<?php 
/**
 * Main handler for the client.
 */

/*
 * Valid combinations of HTTP verb and URI. Code iterates through these and takes the FIRST match, so
 * put the longest ones first.
 */
$validPatterns=array(
		//get by property, e.g, /plate/barcode/90a6
		'GET /[A-Za-z]+/name/[A-Za-z0-9]+'=>'getByName',
		
		//get by property, e.g, /plate/barcode/90a6
		'GET /[A-Za-z]+/[A-Za-z0-9]+/[A-Za-z0-9]+'=>'getByProperty',
		
		//create new record, e.g., /plate/create
		'GET /[A-Za-z]+/create'=>'create',
		
		//get by database ID, e.g, /plate/1234
		'GET /[A-Za-z]+/[0-9]+'=>'getById',

		//get all, e.g., /plate - pagination in request parameters
		'GET /[A-Za-z]+'=>'getAll',
);

//page defaults
$hasConfig=false;

//define class autoloader
spl_autoload_register(function($className){
	$paths=array(
			'../classes/',
			'../classes/core/',
			'../classes/core/exception/',
			'../classes/core/authentication/',
			'../classes/core/interface/',
			'../classes/model/',
	);
	foreach($paths as $path){
		if(file_exists($path.$className.'.class.php')){
			include_once($path.$className.'.class.php');
		}
	}
});

//determine request method and URI
$wwwroot=rtrim($_SERVER['DOCUMENT_ROOT'],'/');
$method=$_SERVER['REQUEST_METHOD'];
$uri=rtrim($_SERVER['REQUEST_URI'],'/');
$uriArgs=explode('/', substr($uri,1));
$req=$method.' '.$uri;

//get request parameters
$parameters=array();
if('GET'==$method){
	$parameters=$_GET;
} else {
	//Shouldn't happen in the client
	//parse_str(file_get_contents("php://input"),$parameters);
	exit;
}

//connect to the database
database::connect();

session::init(new PhpSession());
if(!session::isLoggedIn()){
	include 'templates/core/login/login.php';
	exit;
}

//If no arguments, skip pattern matching and show the homepage
if(0==count($uriArgs) || ''==$uriArgs[0]){
	$objectType="homepage";
	include $wwwroot.'/client/templates/core/homepage/homepage.php';
	exit;
} else if(session::isAdmin() && 2<=count($uriArgs) && 'defaulthomepage'==$uriArgs[1]){
	$isDefaultHomepage=true;
	include $wwwroot.'/client/templates/core/homepage/homepage.php';
	exit;
}

//Now get the appropriate page
//do action
$objectType=null;
$action=null;

$keys=array_keys($validPatterns);
$numPatterns=count($keys);
$uriMatchedPattern=false;
for($i=0;$i<$numPatterns;$i++){
	$pattern=$keys[$i];
	if(preg_match('|^'.$pattern.'/?$|', $req)){
		$action=$validPatterns[$pattern];
		$objectType=$uriArgs[0];
		$uriMatchedPattern=true;
	}
}



try {
	if(!$uriMatchedPattern){
		throw new BadRequestException('Request '.$uri.' did not match a recognised pattern');
	}
	//echo $action."!".$objectType."!";
	if(!$action || !class_exists($objectType) ){
		throw new BadRequestException('Class or method does not exist');
	}
	$helpTexts=forward_static_call_array(array($objectType, 'getFieldHelpTexts'),array());
	$fieldValidations=forward_static_call_array(array($objectType, 'getFieldValidations'),array());
	$validationPatterns=validator::getValidationPatterns();
	switch($action){
		case 'getById':
			$id=$uriArgs[1];
			$args=array($id, $parameters);
			$data=sanitize(forward_static_call_array(array($objectType, 'getById'), $args));
			if(!$data){ throw new NotFoundException('That item does not exist, or you do not have permission to see it'); }
			$canUpdate=forward_static_call_array(array($objectType, 'canUpdate'), $args);
			$pageType='view';
			break;
		case 'getByName':
			$args=array(urldecode($uriArgs[2]), $parameters);
			$data=sanitize(forward_static_call_array(array($objectType, 'getByName'), $args));
			$canUpdate=forward_static_call_array(array($objectType, 'canUpdate'), $args);
			$pageType='view';
			break;
		case 'getByProperty':
			$args=array($uriArgs[1],urldecode($uriArgs[2]), $parameters);
			$data=sanitize(forward_static_call_array(array($objectType, $action), $args));
			$pageType='list';
			break;
		case 'getAll':
			$args=array($parameters);
			$data=sanitize(forward_static_call_array(array($objectType, $action), $args));
			$pageType='list';
			break;
		case 'create':
			$args=array();
			$canCreate=forward_static_call_array(array($objectType, 'canCreate'), $args);
			if(!$canCreate){
				throw new ForbiddenException('You do not have permission to create that.');
			}
			$pageType='create';
			break;
	}
	if(file_exists('templates/core/'.$objectType.'/'.$pageType.'.php')){
		include_once 'templates/core/'.$objectType.'/'.$pageType.'.php';
	} else {
		include_once 'templates/model/'.$objectType.'/'.$pageType.'.php';
	}
	
} catch(Exception $ex){
	$responseCode=$ex->getCode();
	http_response_code($responseCode);
//	include $wwwroot.'/client/templates/core/_error/'.$responseCode.'.php';
	include 'templates/core/_error/'.$responseCode.'.php';
}

function sanitize($data){
	if(is_array($data)){
		foreach($data as $k=>&$v){
			if(is_array($v)){
				$v=sanitize($v);
			} else {
				$v=htmlspecialchars($v);
			}
		}
	}
	return $data;
}


function requireAdmin(){
	if(!session::isAdmin()){
		throw new ForbiddenException('This is for administrators only.');
	}
}

function redirectToView($objectType, $id){
	header('Location: /'.$objectType.'/'.$id);
}

?>